/* For proper formating set tabstop=3 */

#include <lpfgall.h>

/* Definitions of constants defining apex states */
#define MONOPODIAL 0
#define SYMPODIAL  1
#define FLOWERING  2

/* Wrappers for time functions defined graphically via timeline editor */

#define MERISTEM_SCALE			1.0*pleiotropic_scale
#define meristem_size(t)		MERISTEM_SCALE*tfunc(MERISTEM_DIAMETER, t)

#define INTERNODE_SCALE 		1.0*pleiotropic_scale
#define internode_length(t)	INTERNODE_SCALE*tfunc(INTERNODE_LENGTH, t)

#define BRACT_SCALE				1.0*pleiotropic_scale
#define bract_size(t)			BRACT_SCALE*tfunc(BRACT_SIZE, t)

#define PETIOLE_SCALE			1.0*pleiotropic_scale
#define petiole_width(t)		PETIOLE_SCALE*tfunc(PETIOLE_WIDTH, t)

#define FLOWER_SCALE				0.08*pleiotropic_scale
#define flower_size(t)			FLOWER_SCALE*tfunc(FLOWER_DIAMETER, t)

#define COLOR_SCALE				32
#define flower_color(t)			COLOR_SCALE*tfunc(FLOWER_COLOR, t)

/* Declarations of data structures, modules, and global variables */
struct MeristemData {
	int state;	// MONOPODIAL, SYMPODIAL, or FLOWERING
	float age;	// age since creation or production of the most recent primordium
	float d;		// diameter
};
module Meristem(MeristemData);

struct InternodeData {
	float age;		// age since the time of creation
	float v;			// veg at time of creation
	float dn;		// diameter raised to power n (pipe_model_power)
};
module Internode(InternodeData);

struct BractData {
	float age;		// age from the time of creation
	float v;			// veg at time of creation
	float d;			// diameter
};
module Bract(BractData);

module INIT();		// initializes variables used in the definition of initial structrue
module Primordium(float,float); // (age, veg difference between sibling primordia)

float max_age;				// inflorescence age at the end of simulation
float dt;					// simulation time step

float plastochron; 		// plastochron
float delay_mono;			// delay in the development of lateral primordia - raceme
float delay_sym;			// delay in the development of lateral primordia - cyme 
float delay_diff;			// difference in delay between cyme branches

float veg;					// vegetativeness 
float veg_init; 			// initial veg value (usually 1)
float veg_half_life;		// time of veg decline by to 1/2 ofveg_init 
float veg_decline_rate;	// computed from veg_half_life 

float th_m;					// threshold veg for the monopodial meristem switch to flowering
float th_s;					// threshold veg for arresting sympodial meristem production
float th_diff;				// veg difference vetween sybling primordia 

float b_th_1;				// bract threshold for switch from 3 lobes to 1 lobe
float b_th_2;				// bract threshold for switch from 1 lobe to narrow form

float pleiotropic_scale;		// controls overall inflorewcence size
float peduncle_length_ctrl;	// controls peduncle length by adjusting its veg
float int_len_power;		// controls internode length as a power function of veg
float	internode_segments;		// number of segments subdividing an internode

float distal_diameter;	// meristem diameter for pipe model
float pipe_model_power;	// power for propagating internode diameters in the pipe model

float angle_main;			// deflection of the apex direction in the primary axis
float angle_lateral;		// branching angle - monopodial
float angle_sympodial;	// branching angle - sympodial

float int_elasticity; 			// internode "elasticity" 
float int_elasticity_power;	// controls internode elasticity as a power of veg

float bract_angle_branch;		// orients bract supporting lateral branch
float bract_angle_no_branch;	// orients bract not supporting a lateral branch

float ang_adjustment_1;	// branching angle adjustment for petiole supporting 3-lobed bract
float ang_adjustment_2;	// branching angle adjustment for petiole supportinf 1-lobe bract
float	petiole_length_1;	// target petiole length for bracts with 3 lobes
float	petiole_length_2;	// targetpetiole length for bracts with one love 
float petiole_elasticity_1; 	// petiole elasticity for bracts with 3 loges
float petiole_elasticity_2;	// petiole Valishirinelasticity for bracts with 1 lobe

float bract_size_power;	// controls the size of small bracts as a power function of veg

Start: 
{ 
	/* Input parameter values from the parameters.vset file */

	// simulation timing
	max_age = val(MAX_AGE);
	dt = val(DT);

	// branch initiation	
	plastochron = val(PLASTOCHRON);
	delay_mono = val(LATERAL_DELAY_MONO);
	delay_sym = val(LATERAL_DELAY_SYM);
	delay_diff = val(LATERAL_DELAY_DIFF);

	// veg control
	veg_init = val(VEG_INIT);
		veg = veg_init;
	veg_half_life = val(VEG_HALF_LIFE);
		veg_decline_rate = log(2) / veg_half_life;

	// branching thresholds
	th_m = val(TH_MONO);
	th_s = val(TH_SYM);
	th_diff = val(TH_DIFF);

	// bract type thresholds
	b_th_1 = val(BRACT_TH_1);
	b_th_2 = val(BRACT_TH_2);

	// dimensions
	pleiotropic_scale = val(PLEIOTROPIC_SCALE);
	peduncle_length_ctrl = val(PEDUNCLE_LENGTH_CTRL);
   int_len_power = val(INT_LEN_POWER);
	internode_segments = val(INTERNODE_SEGMENTS);

	// pipe model
	distal_diameter = val(DISTAL_DIAMETER);
	pipe_model_power = val(PIPE_MODEL_POWER);

	// branching angles
	angle_main = val(ANGLE_MAIN);
	angle_lateral = val(ANGLE_LATERAL);
	angle_sympodial = val(ANGLE_SYMPODIAL);

	// gravity and tropism
	int_elasticity = val(INT_ELASTICITY);
	int_elasticity_power = val(INT_ELASTICITY_POWER);

	// bracts
	bract_angle_branch = val(BRACT_ANGLE_BRANCH);
	bract_angle_no_branch = val(BRACT_ANGLE_NO_BRANCH);

	ang_adjustment_1 = val(ANG_ADJUSTMENT_1);
	ang_adjustment_2 = val(ANG_ADJUSTMENT_2);
	petiole_length_1 = val(PETIOLE_LENGTH_1);
	petiole_length_2 = val(PETIOLE_LENGTH_2);
	petiole_elasticity_1 = val(PETIOLE_ELASTICITY_1);
	petiole_elasticity_2 = val(PETIOLE_ELASTICITY_2);

   bract_size_power = val(BRACT_SIZE_POWER);

	/* Scan string backward as needed for the pipe model */
	Backward(); 
}

EndEach: 
{
	veg -= veg*veg_decline_rate*dt;
}

consider: Meristem Internode Bract ;	// modules considered in pipe model information flow

derivation length: (int) (max_age / dt);

Axiom: INIT();

/* Define, initialize and produce the initial structure */ 
INIT():
{
	InternodeData InitInternodeData;
	InitInternodeData.age = 0;
	InitInternodeData.v = veg_init*peduncle_length_ctrl;
	InitInternodeData.dn = 0;

	MeristemData InitMeristemData;
	InitMeristemData.state = MONOPODIAL;
	InitMeristemData.age = 0;
	InitMeristemData.d = 0;

	produce Internode(InitInternodeData) Meristem(InitMeristemData);
}

/* Advance meristem age by time step dt, update meristem diameter */
Meristem(m):
{
	m.age += dt; 
 	m.d = fmax(m.d, distal_diameter * meristem_size(m.age));
	produce Meristem(m);
}

/* Advance bract age by time step dt, update petiole diameter */
Bract(b):
{
	b.age += dt; 
	b.d = distal_diameter * petiole_width(b.age);
	produce Bract(b);
}

/* Advance internode age by time step dt, update diameter according to the pipe model */
Internode(s): 
{
	s.age += dt;
	s.dn = 0;

	struct MeristemData mr;		// possible contexts : meristem...
	struct InternodeData sr;	// ...an internode
	struct BractData br;			// ...or a bract

	while (1) { // Sum up all contributions to the width of the supporting internode

		if (InNewRightContext(SB SB Bract(br) EB Internode(sr) EB)) {	// bract and supported branch
			s.dn += pow(br.d,pipe_model_power) + sr.dn;
			continue;
		}

		if (InNewRightContext(SB SB Bract(br) EB EB)) {	// bract not supporting a branch
			s.dn = pow(br.d,pipe_model_power);
			continue;
		}

		if (InNewRightContext(Meristem(mr))) {		// meristem
 			s.dn = pow(mr.d,pipe_model_power);
			break;
		}
		if (InNewRightContext(Internode(sr))) {	// child internode in current axis
			s.dn += sr.dn;
			break;
		}
	}

	produce Internode(s);
}

/* Main part of the model - see the object description */
decomposition:
maximum depth: 5;

Meristem(m):
{
	if (m.age >= plastochron) {

		if (m.state == MONOPODIAL) {	// 1st order meristem: monopodial branching
			if (veg > th_m) {				// high veg: produce a lateral meristem 
				m.age -= plastochron;
				InternodeData s = {m.age, veg ,0};
				nproduce SB Left(angle_lateral) Primordium(m.age-delay_mono, veg-th_s) EB;
				produce Right(angle_main) RollL(180) Internode(s) Meristem(m);
			}
			else {
				m.state = FLOWERING;		// low veg: change state to flowering 
				produce Meristem(m);
			}
		}

		if (m.state == SYMPODIAL)  {	// meristem of order 2 or higher: sympodial branching
			m.age -= plastochron;
			InternodeData s = {m.age, veg, 0};
			nproduce SB Left(angle_sympodial) Primordium(m.age-delay_sym, veg-th_s) EB;
			nproduce SB RollL(180) Left(angle_sympodial) 
				Primordium(m.age-delay_sym-delay_diff, veg-th_s-th_diff) EB;
			m.state = FLOWERING;			// parent meristem always switches to flowering
			produce Internode(s) Meristem(m);
		}
	}
}
		
Primordium(p_age,p_veg) :
{
	BractData b = {p_age, veg, (float) (bract_size(b.age))};
	if (p_veg > 0) {									// If veg is high enough
		MeristemData new_m = {SYMPODIAL, p_age};
		InternodeData s = {new_m.age, veg, 0};
		nproduce SB Left(bract_angle_branch) Bract(b) EB; 	// produce a bract
		produce Internode(s) Meristem(new_m);	// and a meristem supported by an internode
	}
	else {												// If veg is not high enough
		nproduce SB Right(bract_angle_no_branch) Bract(b) EB;	// produce a bract 
		produce;											// and abort
	}
}
		
interpretation:

/* Visualize meristem in flowering state as a sphere with radius and color depending on age */
Meristem(m) :
{  
	if (m.state == FLOWERING) {
		float t = m.age;
		produce SetColor(16+floor(flower_color(t)))
			f(flower_size(t)) Sphere(flower_size(t));
	}
}

/* Visualize internode */
Internode(s):
{
	// Internode length is a function of age and vigor at the time of creation
	float int_length_veg = internode_length(s.age)*pow(s.v,int_len_power);
	// Internode width is set following the pipe model
	float d = pow(s.dn, 1.0/pipe_model_power);
	nproduce SetWidth(d);
	// Introduce some curvature to internodes, inspired by the gravity / tropisms.
	// The magnitude of this effect depends on vigor (purely for visual purposes).
	nproduce Elasticity ((int_elasticity / pow(s.v, int_elasticity_power)) / internode_segments);
	// To capture this curvature, each internode is divided into segments.
	for (int i = 0; i<internode_segments; i++)
		nproduce F(int_length_veg/internode_segments);
	produce;
}

/* Visualize bract in a manner depending on veg */
Bract(b):
{
	// If veg is high, create a leaf-like bract with three lobes
	if (b.v > b_th_1) { 
		nproduce Left(ang_adjustment_1);
		nproduce SetWidth (b.d);
		nproduce Elasticity(petiole_elasticity_1 / internode_segments);
		float petiole_length = petiole_length_1 * bract_size(b.age);
		for (int i = 0; i<internode_segments; i++)
			nproduce F(petiole_length/internode_segments);
		produce CurrentTexture(0) Surface(0,bract_size(b.age));
	}		

	// If veg is somewhat lower, create a bract with a single lobe
	if (b.v > b_th_2) { 
		nproduce Left(ang_adjustment_2);
		nproduce SetWidth (b.d);
		nproduce Elasticity(petiole_elasticity_2 /internode_segments);
		float petiole_length = petiole_length_2 * bract_size(b.age);
		for (int i = 0; i<internode_segments; i++)
			nproduce F(petiole_length/internode_segments);
		produce CurrentTexture(1) Surface(1,bract_size(b.age));
	}		

	// If veg is even less, create a bract with size depending on veg
	float bract_size_veg = bract_size(b.age)*pow(b.v,bract_size_power);
	produce SetColor(2) RollL(90) Surface(2, bract_size_veg);
}
